from __future__ import annotations

from typing import Any

import pytest
from _pytest.monkeypatch import MonkeyPatch

from ckanext.files import config


class TestDefault:
    def test_default_value(self):
        """Verify that default value does not change accidentally.

        If you are modifying this test, don't forget to add changelog entry!

        """
        assert config.default_storage() == "default"

    @pytest.mark.ckan_config(config.DEFAULT_STORAGE, "test")
    def test_customized(self):
        """Default storage can be changed."""
        assert config.default_storage() == "test"


class TestStorages:
    def test_empty(self, ckan_config: dict[str, Any]):
        """With no customization we have only storage defined by test.ini."""
        assert config.storages() == {
            "default": {
                "type": "files:redis",
                "path": "ckanext:files:test.ckan.net:file_content",
                "redis_url": ckan_config["ckan.redis.url"],
                "override_existing": False,
                "name": "default",
                "supported_types": [],
                "location_transformers": [],
                "max_size": 0,
            },
        }

    def test_customized(self, monkeypatch: MonkeyPatch, ckan_config: dict[str, Any]):
        """Storage configuration grouped by the storage name."""
        patches: list[tuple[str, Any]] = [
            ("default.type", "files:redis"),
            ("test.type", "test"),
            ("test.path", "somepath"),
            ("another.type", "fancy"),
            ("another.location_transformers", ["hello"]),
        ]
        for key, value in patches:
            monkeypatch.setitem(ckan_config, config.STORAGE_PREFIX + key, value)

        storages = config.storages()

        assert storages == {
            "default": {
                "type": "files:redis",
                "path": "ckanext:files:test.ckan.net:file_content",
                "redis_url": ckan_config["ckan.redis.url"],
                "override_existing": False,
                "name": "default",
                "supported_types": [],
                "location_transformers": [],
                "max_size": 0,
            },
            "test": {
                "type": "test",
                "path": "somepath",
            },
            "another": {
                "type": "fancy",
                "location_transformers": ["hello"],
            },
        }

    @pytest.mark.ckan_config(config.STORAGE_PREFIX + "another", "test")
    def test_non_setting(self, ckan_config: dict[str, Any]):
        """Test extra items from settings.

        Only `<PREFIX>.<NAME>.<OPTION>` settings are grouped. If `.<OPTION>`
        part is missing, we ignore this line.
        """
        storages = config.storages()

        assert storages == {
            "default": {
                "type": "files:redis",
                "path": "ckanext:files:test.ckan.net:file_content",
                "redis_url": ckan_config["ckan.redis.url"],
                "override_existing": False,
                "name": "default",
                "supported_types": [],
                "location_transformers": [],
                "max_size": 0,
            },
        }



from __future__ import annotations

import copy
from typing import Any, cast

import sqlalchemy as sa
from werkzeug.utils import secure_filename

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.logic import validate
from ckan.types import Action, Context

from ckanext.files import shared, utils
from ckanext.files.shared import File, Location, Multipart, Owner, TransferHistory

from . import schema


@tk.chained_action
def _chained_action(
    next_action: Action,
    context: Context,
    data_dict: dict[str, Any],
) -> Any:
    return next_action(context, data_dict)


package_create = shared.with_task_queue(_chained_action, "package_create")
package_update = shared.with_task_queue(_chained_action, "package_update")
resource_create = shared.with_task_queue(
    _chained_action,
    "resource_create",
)
resource_update = shared.with_task_queue(
    _chained_action,
    "resource_update",
)
group_create = shared.with_task_queue(_chained_action, "group_create")
group_update = shared.with_task_queue(_chained_action, "group_update")
organization_create = shared.with_task_queue(
    _chained_action,
    "organization_create",
)
organization_update = shared.with_task_queue(
    _chained_action,
    "organization_update",
)
user_create = shared.with_task_queue(_chained_action, "user_create")
user_update = shared.with_task_queue(_chained_action, "user_update")


def _flat_mask(data: dict[str, Any]) -> dict[tuple[Any, ...], Any]:
    result: dict[tuple[Any, ...], Any] = {}

    for k, v in data.items():
        if isinstance(v, dict):
            result.update({(k,) + sk: sv for sk, sv in _flat_mask(v).items()})
        else:
            result[(k,)] = v

    return result


@tk.side_effect_free
@validate(schema.file_search)
def files_file_search(  # noqa: C901, PLR0912, PLR0915
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """Search files.

    /// warning
    This action is not stabilized yet and will change in future.
    ///

    Provides an ability to search files using exact filter by name,
    content_type, size, owner, etc. Results are paginated and returned in
    package_search manner, as dict with `count` and `results` items.

    All columns of File model can be used as filters. Before the search, type
    of column and type of filter value are compared. If they are the same,
    original values are used in search. If type different, column value and
    filter value are casted to string.

    This request produces `size = 10` SQL expression:
    ```sh
    ckanapi action files_file_search size:10
    ```

    This request produces `size::text = '10'` SQL expression:
    ```sh
    ckanapi action files_file_search size=10
    ```

    Even though results are usually not changed, using correct types leads to
    more efficient search.

    Apart from File columns, the following Owner properties can be used for
    searching: `owner_id`, `owner_type`, `pinned`.

    `storage_data` and `plugin_data` are dictionaries. Filter's value for these
    fields used as a mask. For example, `storage_data={"a": {"b": 1}}` matches
    any File with `storage_data` *containing* item `a` with value that contains
    `b=1`. This works only with data represented by nested dictionaries,
    without other structures, like list or sets.

    Experimental feature: File columns can be passed as a pair of operator and
    value. This feature will be replaced by strictly defined query language at
    some point:

    ```sh
    ckanapi action files_file_search size:'["<", 100]' content_type:'["like", "text/%"]'
    ```
    Fillowing operators are accepted: `=`, `<`, `>`, `!=`, `like`

    Args:
        start (int): index of first row in result/number of rows to skip. Default: `0`
        rows (int): number of rows to return. Default: `10`
        sort (str): name of File column used for sorting. Default: `name`
        reverse (bool): sort results in descending order. Default: `False`
        storage_data (dict[str, Any]): mask for `storage_data` column. Default: `{}`
        plugin_data (dict[str, Any]): mask for `plugin_data` column. Default: `{}`
        owner_id (str): show only specific owner id if present. Default: `None`
        owner_type (str): show only specific owner type if present. Default: `None`
        pinned (bool): show only pinned/unpinned items if present. Default: `None`
        completed (bool): use `False` to search incomplete uploads. Default: `True`

    Returns:
        dictionary with `count` and `results`
    """
    tk.check_access("files_file_search", context, data_dict)
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

    if data_dict["completed"]:
        stmt = sa.select(File).outerjoin(
            Owner,
            sa.and_(File.id == Owner.item_id, Owner.item_type == "file"),
        )

        inspector: Any = sa.inspect(File)
    else:
        stmt = sa.select(Multipart).outerjoin(
            Owner,
            sa.and_(Multipart.id == Owner.item_id, Owner.item_type == "multipart"),
        )

        inspector: Any = sa.inspect(Multipart)

    for field in ["owner_type", "owner_id", "pinned"]:
        if field in data_dict:
            value = data_dict[field]
            if value is not None and not (
                field == "pinned" and isinstance(value, bool)
            ):
                value = str(value)
            stmt = stmt.where(getattr(Owner, field) == value)

    columns = inspector.columns

    for mask in ["storage_data", "plugin_data"]:
        if mask in data_dict:
            for k, v in _flat_mask(data_dict[mask]).items():
                field = columns[mask]
                for segment in k:
                    field = field[segment]

                stmt = stmt.where(field.astext == v)

    for k, v in data_dict.get("__extras", {}).items():
        if k not in columns:
            continue

        if (
            isinstance(v, list)
            and len(v) == 2  # noqa: PLR2004
            and v[0] in ["=", "<", ">", "!=", "like"]
        ):
            op, v = cast("list[Any]", v)  # noqa: PLW2901
        else:
            op = "="

        col = columns[k]
        column_type = col.type.python_type
        if not isinstance(v, column_type) and v is not None:
            v = str(v)  # noqa: PLW2901
            col = sa.func.cast(col, sa.Text)

        if v is None:
            if op == "=":
                op = "is"
            elif op == "!=":
                op = "is not"

        stmt = stmt.where(col.bool_op(op)(v))

    total = sess.scalar(stmt.with_only_columns(sa.func.count()))

    parts = data_dict["sort"].split(".")
    sort = parts[0]
    sort_path = parts[1:]

    if sort not in columns:
        raise tk.ValidationError({"sort": ["Unknown sort column"]})

    column = columns[sort]

    if sort_path and sort == "storage_data":
        for part in sort_path:
            column = column[part]

    if data_dict["reverse"]:
        column = column.desc()

    stmt = stmt.order_by(column)

    stmt = stmt.limit(data_dict["rows"]).offset(data_dict["start"])

    cache = utils.ContextCache(context)
    results: list[File] = [cache.set("file", f.id, f) for f in sess.scalars(stmt)]
    return {"count": total, "results": [f.dictize(context) for f in results]}


@validate(schema.file_create)
def files_file_create(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Create a new file.

    This action passes uploaded file to the storage without strict
    validation. File is converted into standard upload object and everything
    else is controlled by storage. The same file may be uploaded to one storage
    and rejected by other, depending on configuration.

    This action is way too powerful to use it directly. The recommended
    approach is to register a different action for handling specific type of
    uploads and call current action internally.

    When uploading a real file(or using `werkqeug.datastructures.FileStorage`),
    name parameter can be omited. In this case, the name of uploaded file is used.

    ```sh
    ckanapi action files_file_create upload@path/to/file.txt
    ```

    When uploading a raw content of the file using string or bytes object, name
    is mandatory.

    ```sh
    ckanapi action files_file_create upload@<(echo -n "hello world") name=file.txt
    ```

    Requires storage with `CREATE` capability.

    Args:
        name (str, optional): human-readable name of the file.
            Default: guess using upload field
        storage (str, optional): name of the storage that will handle the upload.
            Default: `default`
        upload (shared.types.Uploadable): content of the file as bytes,
            file descriptor or uploaded file

    Returns:
        dictionary with file details.
    """
    tk.check_access("files_file_create", context, data_dict)
    extras = data_dict.get("__extras", {})

    try:
        storage = shared.get_storage(data_dict["storage"])
    except shared.exc.UnknownStorageError as err:
        raise tk.ValidationError({"storage": [str(err)]}) from err

    if not storage.supports(shared.Capability.CREATE):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    filename = secure_filename(data_dict["name"])

    try:
        storage_data = storage.upload(
            storage.prepare_location(filename, data_dict["upload"]),
            data_dict["upload"],
            **extras,
        )
    except (shared.exc.UploadError, shared.exc.ExistingFileError) as err:
        raise tk.ValidationError({"upload": [str(err)]}) from err

    fileobj = File(
        name=filename,
        storage=data_dict["storage"],
    )
    storage_data.into_object(fileobj)
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    sess.add(fileobj)

    _set_user_owner(context, "file", fileobj.id)
    sess.commit()

    utils.ContextCache(context).set("file", fileobj.id, fileobj)

    return fileobj.dictize(context)


@shared.with_task_queue
@validate(schema.file_replace)
def files_file_replace(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Replace content of the file.

    New file must have the same MIMEtype as the original file.

    Size and content hash from the new file will replace original values. All
    other fields, including name, remain unchanged.

    ```sh
    ckanapi action files_file_replace id=123 upload@path/to/file.txt
    ```

    Requires storage with `CREATE` and `REMOVE` capability.

    Args:
        id (str): ID of the replaced file
        upload (shared.types.Uploadable): content of the file as bytes,
            file descriptor or uploaded file

    Returns:
        dictionary with file details.

    """
    tk.check_access("files_file_replace", context, data_dict)

    cache = utils.ContextCache(context)

    fileobj = cache.get_model("file", data_dict["id"], File)
    if not fileobj:
        raise tk.ObjectNotFound("file")

    if fileobj.content_type != data_dict["upload"].content_type:
        raise tk.ValidationError(
            {
                "upload": [
                    "Expected {} but received {}".format(
                        fileobj.content_type,
                        data_dict["upload"].content_type,
                    ),
                ],
            },
        )
    storage = shared.get_storage(fileobj.storage)

    if not storage.supports(shared.Capability.CREATE | shared.Capability.REMOVE):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    try:
        storage_data = storage.upload(
            storage.prepare_location(fileobj.name, data_dict["upload"]),
            data_dict["upload"],
        )
    except shared.exc.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]}) from err

    # with transparent location strategy file will be writted into the same
    # location, so removal may be not required
    if storage_data.location != fileobj.location:
        old_data = shared.FileData.from_object(fileobj)
        shared.add_task(lambda result, idx, prev: storage.remove(old_data))

    storage_data.into_object(fileobj)
    context["session"].commit()  # pyright: ignore[reportTypedDictNotRequiredAccess]

    utils.ContextCache(context).set("file", fileobj.id, fileobj)

    return fileobj.dictize(context)


def _set_user_owner(context: Context, item_type: str, item_id: str):
    """Add user from context as file owner."""
    user = model.User.get(context.get("user", ""))
    if user:
        owner = Owner(
            item_id=item_id,
            item_type=item_type,
            owner_id=user.id,
            owner_type="user",
        )
        context["session"].add(owner)  # pyright: ignore[reportTypedDictNotRequiredAccess]


@validate(schema.file_delete)
def files_file_delete(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Remove file from storage.

    Unlike packages, file has no `state` field. Removal usually means that file
    details removed from DB and file itself removed from the storage.

    Some storage can implement revisions of the file and keep archived versions
    or backups. Check storage documentation if you need to know whether there
    are chances that file is not completely removed with this operation.

    Requires storage with `REMOVE` capability.

    ```sh
    ckanapi action files_file_delete id=226056e2-6f83-47c5-8bd2-102e2b82ab9a
    ```

    Args:
        id (str): ID of the file
        completed (bool): use `False` to remove incomplete uploads. Default: `True`

    Returns:
        dictionary with details of the removed file.
    """
    tk.check_access("files_file_delete", context, data_dict)

    cache = utils.ContextCache(context)

    fileobj = cache.get_model(
        "file", data_dict["id"], File if data_dict["completed"] else Multipart
    )

    if not fileobj:
        raise tk.ObjectNotFound("file")

    storage = shared.get_storage(fileobj.storage)
    if not storage.supports(shared.Capability.REMOVE):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    dc = shared.FileData if data_dict["completed"] else shared.MultipartData
    try:
        storage.remove(dc.from_object(fileobj))
    except shared.exc.PermissionError as err:
        raise tk.NotAuthorized(str(err)) from err

    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    sess.delete(fileobj)
    sess.commit()

    return fileobj.dictize(context)


@tk.side_effect_free
@validate(schema.file_show)
def files_file_show(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Show file details.

    This action only displays information from DB record. There is no way to
    get the content of the file using this action(or any other API action).

    ```sh
    ckanapi action files_file_show id=226056e2-6f83-47c5-8bd2-102e2b82ab9a
    ```

    Args:
        id (str): ID of the file
        completed (bool): use `False` to show incomplete uploads. Default: `True`

    Returns:
        dictionary with file details
    """
    tk.check_access("files_file_show", context, data_dict)

    cache = utils.ContextCache(context)
    fileobj = cache.get_model(
        "file", data_dict["id"], File if data_dict["completed"] else Multipart
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    utils.ContextCache(context).set("file", fileobj.id, fileobj)
    return fileobj.dictize(context)


@validate(schema.file_rename)
def files_file_rename(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Rename the file.

    This action changes human-readable name of the file, which is stored in
    DB. Real location of the file in the storage is not modified.

    ```sh
    ckanapi action files_file_show \\
        id=226056e2-6f83-47c5-8bd2-102e2b82ab9a \\
        name=new-name.txt
    ```

    Args:
        id (str): ID of the file
        name (str): new name of the file
        completed (bool): use `False` to rename incomplete uploads. Default: `True`

    Returns:
        dictionary with file details

    """
    tk.check_access("files_file_rename", context, data_dict)

    cache = utils.ContextCache(context)
    fileobj = cache.get_model(
        "file",
        data_dict["id"],
        File if data_dict["completed"] else Multipart,
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    fileobj.name = secure_filename(data_dict["name"])

    context["session"].commit()  # pyright: ignore[reportTypedDictNotRequiredAccess]

    return fileobj.dictize(context)


@validate(schema.multipart_start)
def files_multipart_start(
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """Initialize multipart(resumable,continuous,signed,etc) upload.

    Apart from standard parameters, different storages can require additional
    data, so always check documentation of the storage before initiating
    multipart upload.

    When upload initialized, storage usually returns details required for
    further upload. It may be a presigned URL for direct upload, or just an ID
    of upload which must be used with `files_multipart_update`.

    Requires storage with `MULTIPART` capability.

    Args:
        storage (str): name of the storage that will handle the upload.
            Default: `default`
        name (str): name of the uploaded file.
        content_type (str): MIMEtype of the uploaded file. Used for validation
        size (oint): Expected size of upload. Used for validation
        hash (str): Expected content hash. If present, used for validation.

    Returns:
        dictionary with details of initiated upload. Depends on used storage
    """
    tk.check_access("files_multipart_start", context, data_dict)
    extras = data_dict.get("__extras", {})

    try:
        storage = shared.get_storage(data_dict["storage"])
    except shared.exc.UnknownStorageError as err:
        raise tk.ValidationError({"storage": [str(err)]}) from err

    if not storage.supports(shared.Capability.MULTIPART):
        raise tk.ValidationError({"storage": ["Operation is not supported"]})

    filename = secure_filename(data_dict["name"])
    data = shared.MultipartData(
        Location(filename),
        data_dict["size"],
        data_dict["content_type"],
        data_dict["hash"],
    )

    location = storage.prepare_location(filename, data)
    try:
        data = storage.multipart_start(
            location,
            data,
            **extras,
        )
    except shared.exc.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]}) from err

    fileobj = Multipart(
        name=filename,
        storage=data_dict["storage"],
    )
    data.into_object(fileobj)

    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    sess.add(fileobj)
    _set_user_owner(context, "multipart", fileobj.id)
    sess.commit()

    utils.ContextCache(context).set("file", fileobj.id, fileobj)
    return fileobj.dictize(context)


@validate(schema.multipart_refresh)
def files_multipart_refresh(
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """Refresh details of incomplete upload.

    Can be used if upload process was interrupted and client does not know how
    many bytes were already uploaded.

    Requires storage with `MULTIPART` capability.

    Args:
        id (str): ID of the incomplete upload

    Returns:
        dictionary with details of the updated upload

    """
    tk.check_access("files_multipart_refresh", context, data_dict)

    cache = utils.ContextCache(context)
    fileobj = cache.get_model(
        "file",
        data_dict["id"],
        Multipart,
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    storage = shared.get_storage(fileobj.storage)
    storage.multipart_refresh(shared.MultipartData.from_object(fileobj)).into_object(
        fileobj,
    )
    context["session"].commit()  # pyright: ignore[reportTypedDictNotRequiredAccess]

    return fileobj.dictize(context)


@validate(schema.multipart_update)
def files_multipart_update(
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """Update incomplete upload.

    Depending on storage this action may require additional parameters. Most
    likely, `upload` with the fragment of uploaded file.

    Requires storage with `MULTIPART` capability.

    Args:
        id (str): ID of the incomplete upload

    Returns:
        dictionary with details of the updated upload

    """
    tk.check_access("files_multipart_update", context, data_dict)

    extras = data_dict.get("__extras", {})

    cache = utils.ContextCache(context)
    fileobj = cache.get_model(
        "file",
        data_dict["id"],
        Multipart,
    )
    if not fileobj:
        raise tk.ObjectNotFound("upload")

    storage = shared.get_storage(fileobj.storage)

    try:
        storage.multipart_update(
            shared.MultipartData.from_object(fileobj),
            **extras,
        ).into_object(fileobj)
    except shared.exc.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]}) from err

    context["session"].commit()  # pyright: ignore[reportTypedDictNotRequiredAccess]

    return fileobj.dictize(context)


@validate(schema.multipart_complete)
def files_multipart_complete(
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """Finalize multipart upload and transform it into completed file.

    Depending on storage this action may require additional parameters. But
    usually it just takes ID and verifies that content type, size and hash
    provided when upload was initialized, match the actual value.

    If data is valid and file is completed inside the storage, new File entry
    with file details created in DB and file can be used just as any normal
    file.

    Requires storage with `MULTIPART` capability.

    Args:
        id (str): ID of the incomplete upload
        keep_storage_data (bool): copy storage data from multipart upload
        keep_plugin_data (bool): copy plugin data from multipart upload

    Returns:
        dictionary with details of the created file

    """
    tk.check_access("files_multipart_complete", context, data_dict)
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    extras = data_dict.get("__extras", {})

    cache = utils.ContextCache(context)
    multipart = cache.get_model("file", data_dict["id"], Multipart)
    if not multipart:
        raise tk.ObjectNotFound("upload")

    storage = shared.get_storage(multipart.storage)

    fileobj = File(
        name=multipart.name,
        storage=multipart.storage,
    )

    try:
        storage.multipart_complete(
            shared.MultipartData.from_object(multipart),
            **extras,
        ).into_object(fileobj)
    except shared.exc.UploadError as err:
        raise tk.ValidationError({"upload": [str(err)]}) from err

    if data_dict["keep_storage_data"]:
        data: Any = copy.deepcopy(multipart.storage_data)
        data.update(fileobj.storage_data)
        fileobj.storage_data = data

    if data_dict["keep_plugin_data"]:
        data: Any = copy.deepcopy(multipart.plugin_data)
        data.update(fileobj.plugin_data)
        fileobj.plugin_data = data

    if owner_info := multipart.owner_info:
        sess.add(
            Owner(
                item_id=fileobj.id,
                item_type="file",
                owner_id=owner_info.owner_id,
                owner_type=owner_info.owner_type,
            )
        )

    sess.add(fileobj)
    sess.delete(multipart)
    sess.commit()

    cache.set("file", fileobj.id, fileobj)
    return fileobj.dictize(context)


@tk.side_effect_free
@validate(schema.file_scan)
def files_file_scan(
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """List files of the owner.

    This action internally calls files_file_search, but with static values of
    owner filters. If owner is not specified, files filtered by current
    user. If owner is specified, user must pass authorization check to see
    files.

    Args:
        owner_id (str): ID of the owner
        owner_type (str): type of the owner
        **rest (Any): The all other parameters are passed as-is to `files_file_search`.

    Returns:
        dictionary with `count` and `results`
    """
    if not data_dict["owner_id"] and data_dict["owner_type"] == "user":
        user = context.get("auth_user_obj")

        if isinstance(user, model.User) or (user := model.User.get(context["user"])):  # pyright: ignore[reportTypedDictNotRequiredAccess]
            data_dict["owner_id"] = user.id

    tk.check_access("files_file_scan", context, data_dict)

    params = data_dict.pop("__extras", {})
    params.update(data_dict)

    # TODO: pull cache from search context
    return tk.get_action("files_file_search")({"ignore_auth": True}, params)


@validate(schema.transfer_ownership)
def files_transfer_ownership(
    context: Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    """Transfer file ownership.

    Args:
        id (str): ID of the file upload
        completed (bool): use `False` to transfer incomplete uploads. Default: `True`
        owner_id (str): ID of the new owner
        owner_type (str): type of the new owner
        force (bool): move file even if it's pinned. Default: `False`
        pin (bool): pin file after transfer to stop future transfers. Default: `False`

    Returns:
        dictionary with details of updated file
    """
    tk.check_access("files_transfer_ownership", context, data_dict)
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

    cache = utils.ContextCache(context)

    fileobj = cache.get_model(
        "file",
        data_dict["id"],
        File if data_dict["completed"] else Multipart,
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    owner = fileobj.owner_info
    if not owner:
        owner = Owner(
            item_id=fileobj.id,
            item_type="file" if data_dict["completed"] else "multipart",
        )
        context["session"].add(owner)  # pyright: ignore[reportTypedDictNotRequiredAccess]

    elif owner.pinned and not data_dict["force"]:
        raise tk.ValidationError(
            {
                "force": ["Must be enabled to transfer pinned files"],
            },
        )
    elif (owner.owner_type, owner.owner_id) != (
        data_dict["owner_type"],
        data_dict["owner_id"],
    ):
        archive = TransferHistory.from_owner(owner)
        archive.actor = context["user"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        sess.add(archive)

    owner.owner_id = data_dict["owner_id"]
    owner.owner_type = data_dict["owner_type"]
    owner.pinned = data_dict["pin"]
    sess.expire(fileobj)
    sess.commit()

    return fileobj.dictize(context)


@validate(schema.file_pin)
def files_file_pin(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Pin file to the current owner.

    Pinned file cannot be transfered to a different owner. Use it to guarantee
    that file referred by entity is not accidentally transferred to a different
    owner.

    Args:
        id (str): ID of the file
        completed (bool): use `False` to pin incomplete uploads. Default: `True`

    Returns:
        dictionary with details of updated file
    """
    tk.check_access("files_file_pin", context, data_dict)
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

    cache = utils.ContextCache(context)
    fileobj = cache.get_model(
        "file",
        data_dict["id"],
        File if data_dict["completed"] else Multipart,
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    owner = fileobj.owner_info
    if not owner:
        raise tk.ValidationError({"id": ["File has no owner"]})

    owner.pinned = True
    sess.commit()

    return fileobj.dictize(context)


@validate(schema.file_unpin)
def files_file_unpin(context: Context, data_dict: dict[str, Any]) -> dict[str, Any]:
    """Pin file to the current owner.

    Pinned file cannot be transfered to a different owner. Use it to guarantee
    that file referred by entity is not accidentally transferred to a different
    owner.

    Args:
        id (str): ID of the file
        completed (bool): use `False` to unpin incomplete uploads. Default: `True`

    Returns:
        dictionary with details of updated file
    """
    tk.check_access("files_file_unpin", context, data_dict)
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

    cache = utils.ContextCache(context)
    fileobj = cache.get_model(
        "file",
        data_dict["id"],
        File if data_dict["completed"] else Multipart,
    )
    if not fileobj:
        raise tk.ObjectNotFound("file")

    owner = fileobj.owner_info
    if not owner:
        raise tk.ValidationError({"id": ["File has no owner"]})

    owner.pinned = False
    sess.commit()

    return fileobj.dictize(context)


@validate(schema.resource_upload)
def files_resource_upload(
    context: Context, data_dict: dict[str, Any]
) -> dict[str, Any]:
    """Create a new file inside resource storage.

    This action internally calls `files_file_create` with `ignore_auth=True`
    and always uses resources storage.

    New file is not attached to resource. You need to call
    `files_transfer_ownership` manually, when resource created. Or you can use
    `files_transfer_ownership("resource","id")` validator to do it
    automatically.

    Args:
        name (str): human-readable name of the file.
            Default: guess using upload field
        upload (shared.types.Uploadable): content of the file as bytes,
            file descriptor or uploaded file

    Returns:
        dictionary with file details

    """
    tk.check_access("files_resource_upload", context, data_dict)
    storage_name = shared.config.resources_storage()
    if not storage_name:
        raise tk.ValidationError(
            {"upload": ["Resource uploads are not supported"]},
        )

    # TODO: pull cache from the context
    return tk.get_action(
        "files_multipart_start" if data_dict["multipart"] else "files_file_create"
    )(
        Context(context, ignore_auth=True),
        dict(data_dict, storage=storage_name),
    )



from __future__ import annotations

import os
import sys
from typing import IO

import click

import ckan.plugins.toolkit as tk
from ckan import model

from ckanext.files import shared


@click.group()
def group():
    """File-level operations."""


@group.command()
@click.argument("file_id")
@click.option("--start", type=int, default=0, help="Start streaming from position")
@click.option("--end", type=int, help="End streaming at position")
@click.option("-o", "--output", help="Stream into specified file or directory")
def stream(file_id: str, output: str | None, start: int, end: int | None):
    """Stream content of the file."""
    file = model.Session.get(shared.File, file_id)
    if not file:
        tk.error_shout("File not found")
        raise click.Abort

    try:
        storage = shared.get_storage(file.storage)
    except shared.exc.UnknownStorageError as err:
        tk.error_shout(err)
        raise click.Abort from err

    if not start and end is None and storage.supports(shared.Capability.STREAM):
        content_stream = storage.stream(shared.FileData.from_object(file))

    elif storage.supports(shared.Capability.RANGE):
        content_stream = storage.range(shared.FileData.from_object(file), start, end)

    elif storage.supports_synthetic(shared.Capability.RANGE, storage):
        content_stream = storage.range_synthetic(
            shared.FileData.from_object(file), start, end
        )

    else:
        tk.error_shout("File streaming is not supported")
        raise click.Abort

    if output is None:
        dest: IO[bytes] = sys.stdout.buffer
    else:
        if os.path.isdir(output):
            output = os.path.join(output, file.name)
        dest = open(output, "wb")  # noqa: SIM115

    for chunk in content_stream:
        click.echo(chunk, nl=False, file=dest)



from __future__ import annotations

import dataclasses

from file_keeper.default.adapters import opendal as od

from ckan.config.declaration import Declaration, Key

from ckanext.files import shared


@dataclasses.dataclass()
class Settings(shared.Settings, od.Settings):
    pass


class OpenDalStorage(shared.Storage, od.OpenDalStorage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)
        declaration.declare(key.scheme).required().set_description(
            "OpenDAL service type. Check available services at"
            + "  https://docs.rs/opendal/latest/opendal/services/index.html",
        )
        declaration.declare(key.params).set_description(
            "JSON object with parameters passed directly to OpenDAL operator.",
        ).set_validators("default({}) convert_to_json_if_string dict_only")
        declaration.declare(key.path).set_description(
            "Path inside the container where uploaded data will be stored.",
        )



from __future__ import annotations

import dataclasses
import logging
import os
from typing import Any

import file_keeper as fk
import flask
from file_keeper.default.adapters import fs

from ckan import types
from ckan.config.declaration import Declaration, Key

from ckanext.files import shared

log = logging.getLogger(__name__)
CHUNK_SIZE = 16384


@dataclasses.dataclass()
class Settings(shared.Settings, fs.Settings):
    pass


class FsStorage(shared.Storage, fs.FsStorage):
    """Store files in local filesystem."""

    settings: Settings  # type: ignore
    SettingsFactory = Settings

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)
        declaration.declare(key.path).required().set_description(
            "Path to the folder where uploaded data will be stored.",
        )

        declaration.declare_bool(key.create_path).set_description(
            "Create storage folder if it does not exist.",
        )

        declaration.declare_bool(key.recursive).set_description(
            "Use this flag if files can be stored inside subfolders"
            + " of the main storage path.",
        )

    def _base_response(
        self, data: fk.FileData, extras: dict[str, Any]
    ) -> types.Response:
        filepath = os.path.join(self.settings.path, data.location)
        return flask.send_file(
            filepath,
            download_name=data.location,
            mimetype=data.content_type,
        )


class PublicFsReader(fs.Reader):
    capabilities = fs.Reader.capabilities | fk.Capability.PERMANENT_LINK
    storage: PublicFsStorage  # type: ignore

    def permanent_link(self, data: fk.FileData, extras: dict[str, Any]) -> str:
        """Return public download link."""
        from ckan.lib.helpers import url_for_static

        return url_for_static(
            os.path.join(
                self.storage.settings.public_prefix,
                data.location,
            ),
            _external=True,
        )


@dataclasses.dataclass()
class PublicFsSettings(Settings):
    public_prefix: str = "/"


class PublicFsStorage(FsStorage):
    settings: PublicFsSettings  # type: ignore
    SettingsFactory = PublicFsSettings
    ReaderFactory = PublicFsReader

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)

        declaration.declare(key.public_prefix).set_description(
            "URL of the storage folder."
            + " `public_prefix + location` must produce a public URL",
        )


class CkanResourceFsStorage(FsStorage):
    def __init__(self, settings: Any):
        settings["recursive"] = True
        super().__init__(settings)



from __future__ import annotations

import dataclasses
from typing import Any

from file_keeper.default.adapters import libcloud as lc

from ckan.config.declaration import Declaration, Key

from ckanext.files import shared

PROVIDERS_URL = (
    "https://libcloud.readthedocs.io/en/stable/storage/"
    + "supported_providers.html#provider-matrix"
)

get_driver: Any


@dataclasses.dataclass()
class Settings(shared.Settings, lc.Settings):
    pass


class LibCloudStorage(shared.Storage, lc.LibCloudStorage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)
        declaration.declare(key.provider).required().set_description(
            "apache-libcloud storage provider. List of providers available at"
            + f" {PROVIDERS_URL} . Use upper-cased value from Provider Constant column",
        )
        declaration.declare(key.key).required().set_description("API key or username")
        declaration.declare(key.secret).set_description("Secret password")
        declaration.declare(key.container_name).required().set_description(
            "Name of the container(bucket)",
        )
        declaration.declare(key.params).set_description(
            "JSON object with additional parameters"
            + " passed directly to storage constructor.",
        ).set_validators("default({}) convert_to_json_if_string dict_only")
        declaration.declare(key.path).set_description(
            "Path inside the container where uploaded data will be stored.",
        )

        declaration.declare(key.public_prefix).set_description(
            "URL prefix to use when builing public file's URL. Usually, this "
            "requires a container with public URL. For example, if storage "
            "uses cloud provider `example.cloud.com` and files are uploaded "
            "into container `my_files`, the public prefix must be set to "
            "`https://certain.cloud.com/my_files`, assuming container can be "
            "anonymously accessed via this URL. File location will be appended"
            " to public prefix, producing absolute public URL."
        )



from __future__ import annotations

from typing import Any

import pytest
from faker import Faker

from ckan import model

from ckanext.files.model import File, Owner


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFile:
    def test_cascade_owner(self, user: dict[str, Any], faker: Faker):
        file = File(
            name=faker.file_name(),
            storage="default",
            location=faker.file_name(),
        )
        owner = Owner(
            item_id=file.id,
            item_type="file",
            owner_id=user["id"],
            owner_type="user",
        )

        model.Session.add_all([file, owner])
        model.Session.commit()
        assert file.owner_info is owner
        assert isinstance(file.owner, model.User)

        model.Session.delete(owner)
        model.Session.commit()

        assert model.Session.get(File, file.id)

        owner = Owner(
            item_id=file.id,
            item_type="file",
            owner_id=user["id"],
            owner_type="user",
        )
        model.Session.add(owner)
        model.Session.commit()
        model.Session.refresh(file)

        assert file.owner_info is owner

        model.Session.delete(file)
        model.Session.commit()
        assert not model.Session.get(Owner, (owner.item_id, owner.item_type))



from __future__ import annotations

import dataclasses

from file_keeper.default.adapters import redis as rd

import ckan.plugins.toolkit as tk
from ckan.config.declaration import Declaration, Key

from ckanext.files import shared


def _default_prefix():
    return "ckanext:files:{}:file_content".format(tk.config["ckan.site_id"])


@dataclasses.dataclass()
class Settings(shared.Settings, rd.Settings):
    pass


class RedisStorage(shared.Storage, rd.RedisStorage):
    settings: Settings  # type: ignore
    SettingsFactory = Settings

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)
        declaration.declare(key.path, _default_prefix()).set_description(
            "Name of the Redis key for HASH with files.",
        )
        declaration.declare(key.redis_url).set_description(
            "Static prefix of the Redis key generated for every upload.",
        ).append_validators("configured_default('ckan.redis.url',None)")



"""Base abstract functionality of the extentsion.

All classes required for specific storage implementations are defined
here. Some utilities, like `make_storage` are also added to this module instead
of `utils` to avoid import cycles.

This module relies only on types, exceptions and utils to prevent import
cycles.

"""

from __future__ import annotations

import dataclasses
from time import time
from typing import Any

import file_keeper as fk
import flask
from typing_extensions import TypeAlias

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.config.declaration import Declaration, Key

from . import config, utils

adapters = fk.adapters
storages = fk.Registry["fk.Storage"]()
Uploader: TypeAlias = fk.Uploader
Manager: TypeAlias = fk.Manager
Reader: TypeAlias = fk.Reader

FileData: TypeAlias = fk.FileData
MultipartData: TypeAlias = fk.MultipartData

make_storage = fk.make_storage


def get_storage(name: str | None = None) -> Storage:
    """Return existing storage instance.

    Storages are initialized when plugin is loaded. As result, this function
    always returns the same storage object for the given name.

    If no name specified, default storage is returned.

    Args:
        name: name of the configured storage

    Returns:
        storage instance

    Raises:
        UnknownStorageError: storage with the given name is not configured

    Example:
        ```
        default_storage = get_storage()
        storage = get_storage("storage name")
        ```

    """
    if name is None:
        name = config.default_storage()

    storage = storages.get(name)

    if not storage:
        raise fk.exc.UnknownStorageError(name)

    return storage  # type: ignore


@dataclasses.dataclass()
class Settings(fk.Settings):
    supported_types: list[str] = dataclasses.field(default_factory=list)
    max_size: int = 0


class Storage(fk.Storage):
    """Base class for storage implementation."""

    settings: Settings  # type: ignore
    SettingsFactory = Settings

    def validate_size(self, size: int):
        max_size = self.settings.max_size
        if max_size and size > max_size:
            raise fk.exc.LargeUploadError(size, max_size)

    def validate_content_type(self, content_type: str):
        supported_types = self.settings.supported_types
        if supported_types and not utils.is_supported_type(
            content_type,
            supported_types,
        ):
            raise fk.exc.WrongUploadTypeError(content_type)

    def upload(
        self, location: fk.Location, upload: fk.Upload, /, **kwargs: Any
    ) -> FileData:
        self.validate_size(upload.size)
        self.validate_content_type(upload.content_type)

        return super().upload(location, upload, **kwargs)

    def multipart_start(
        self, location: fk.Location, data: MultipartData, /, **kwargs: Any
    ) -> MultipartData:
        self.validate_size(data.size)
        self.validate_content_type(data.content_type)

        return super().multipart_start(location, data, **kwargs)

    def temporal_link(self, data: FileData, /, **kwargs: Any) -> str:
        try:
            link = super().temporal_link(data, **kwargs)
        except fk.exc.UnsupportedOperationError:
            link = None

        if not link:
            token = utils.encode_token(
                {
                    "topic": "download_file",
                    "exp": str(int(time()) + kwargs.get("ttl", 30)),
                    "storage": self.settings.name,
                    "location": data.location,
                },
            )
            link = tk.url_for("files.temporal_download", token=token, _external=True)
        return link

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        declaration.declare(key.max_size, 0).append_validators(
            "files_parse_filesize",
        ).set_description(
            "The maximum size of a single upload."
            + "\nSupports size suffixes: 42B, 2M, 24KiB, 1GB."
            + " `0` means no restrictions.",
        )

        declaration.declare_list(key.supported_types, None).set_description(
            "Space-separated list of MIME types or just type or subtype part."
            + "\nExample: text/csv pdf application video jpeg",
        )

        declaration.declare_bool(key.override_existing).set_description(
            "If file already exists, replace it with new content.",
        )

        declaration.declare(key.name, key[-1]).set_description(
            "Descriptive name of the storage used for debugging. When empty,"
            + " name from the config option is used,"
            + " i.e: `ckanext.files.storage.DEFAULT_NAME...`",
        )

        declaration.declare_list(key.location_transformers, None).set_description(
            "List of transformations applied to the file location."
            " Depending on the storage type, sanitizing the path or removing"
            " special characters can be sensible. Empty value leaves location"
            " unchanged, `uuid` transforms location into UUID, `uuid_with_extension`"
            " transforms filename into UUID and appends original file's extension"
            " to it.",
        )

    def as_response(
        self,
        data: FileData,
        filename: str | None = None,
        /,
        send_inline: bool = False,
        **kwargs: Any,
    ) -> types.Response:
        """Make Flask response with file attachment.

        By default, files are served as attachments and are downloaded as
        result.

        If rendering is safe and preferable enable ``send_inline`` flag.

        Args:
            data: file details
            filename: expected name of the file used instead of the real name

        Keyword Args:
            send_inline: do not force download and try rendering file in browser
            **kwargs: ...

        Returns:
            Flask response with file's content
        """
        resp = self._base_response(data, kwargs)

        inline_types = config.inline_types()
        disposition = (
            "inline"
            if send_inline or utils.is_supported_type(data.content_type, inline_types)
            else "attachment"
        )

        resp.headers.set(
            "content-disposition",
            disposition,
            filename=filename or data.location,
        )

        return resp

    def _base_response(self, data: FileData, extras: dict[str, Any]) -> types.Response:
        if not self.supports(fk.Capability.STREAM):
            raise fk.exc.UnsupportedOperationError("stream", self)

        return flask.Response(
            self.reader.stream(data, extras),
            mimetype=data.content_type or None,
        )



from __future__ import annotations

from typing import Any, Iterable

import click
import sqlalchemy as sa

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.search import rebuild

from ckanext.files import shared


@click.group()
def group():
    """Migrate from original CKAN implementation."""


@group.command("groups")
@click.argument("storage_name")
def migrate_groups(storage_name: str):
    """Migrate group images to specified storage."""
    content = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": 0},
    )
    if not content["count"]:
        tk.error_shout(f"Storage {storage_name} contains 0 files.")
        tk.error_shout("Make sure it points to directory with group images and run:")
        tk.error_shout(f"\tckan files scan -s {storage_name} -t")
        raise click.Abort

    click.echo(f"Found {content['count']} files. Searching file owners...")

    files = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": content["count"]},
    )["results"]

    unowned: list[dict[str, Any]] = []
    owned: dict[tuple[str, bool], dict[str, Any]] = {}

    bar: Iterable[Any]
    with click.progressbar(files) as bar:
        for info in bar:
            group = model.Session.scalar(
                sa.select(model.Group).where(model.Group.image_url == info["location"]),
            )

            if group:
                owned[(group.id, group.is_organization)] = info
            else:
                unowned.append(info)

    click.echo(f"Located owners for {len(owned)} files out of {content['count']}.")
    if click.confirm("Show group IDs and corresponding file?"):
        for key, info in owned.items():
            click.echo(f"{key[0]}: {info['location']}")

    if unowned and click.confirm("Show files that do not belong to any group?"):
        for info in unowned:
            click.echo(f"{info['location']}")

    if click.confirm("Transfer file ownership to group identified in previous steps?"):
        with click.progressbar(owned.items()) as bar:
            for key, info in bar:
                group_id, is_organization = key
                bar.label = f"Transfering {info['location']}"
                tk.get_action("files_transfer_ownership")(
                    {"ignore_auth": True},
                    {
                        "pin": True,
                        "force": True,
                        "id": info["id"],
                        "owner_type": "organization" if is_organization else "group",
                        "owner_id": group_id,
                    },
                )


@group.command("users")
@click.argument("storage_name")
def migrate_users(storage_name: str):
    """Migrate user avatars to specified storage."""
    content = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": 0},
    )
    if not content["count"]:
        tk.error_shout(f"Storage {storage_name} contains 0 files.")
        tk.error_shout("Make sure it points to directory with user images and run:")
        tk.error_shout(f"\tckan files scan -s {storage_name} -t")
        raise click.Abort

    click.echo(f"Found {content['count']} files. Searching file owners...")

    files = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": content["count"]},
    )["results"]

    unowned: list[dict[str, Any]] = []
    owned: dict[str, dict[str, Any]] = {}

    bar: Iterable[Any]
    with click.progressbar(files) as bar:
        for info in bar:
            user = model.Session.scalar(
                sa.select(model.User).where(model.User.image_url == info["location"]),
            )

            if user:
                owned[user.id] = info

            else:
                unowned.append(info)

    click.echo(f"Located owners for {len(owned)} files out of {content['count']}.")
    if click.confirm("Show user IDs and corresponding file?"):
        for key, info in owned.items():
            click.echo(f"{key}: {info['location']}")

    if unowned and click.confirm("Show files that do not belong to any user?"):
        for info in unowned:
            click.echo(f"{info['location']}")

    if click.confirm("Transfer file ownership to users identified in previous steps?"):
        with click.progressbar(owned.items()) as bar:
            for user_id, info in bar:
                bar.label = f"Transfering {info['location']}"
                tk.get_action("files_transfer_ownership")(
                    {"ignore_auth": True},
                    {
                        "pin": True,
                        "force": True,
                        "id": info["id"],
                        "owner_type": "user",
                        "owner_id": user_id,
                    },
                )


@group.command("local-resources")
@click.argument("storage_name")
def migrate_local_resources(storage_name: str):  # noqa: C901, PLR0915, PLR0912
    """Migrate resources uploaded via original ResourceUploader."""
    content = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": 0},
    )
    if not content["count"]:
        tk.error_shout(f"Storage {storage_name} contains 0 files.")
        tk.error_shout("Make sure it points to directory with user images and run:")
        tk.error_shout(f"\tckan files scan -s {storage_name} -t")
        raise click.Abort

    click.echo(f"Found {content['count']} files. Searching file owners...")

    files = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": content["count"]},
    )["results"]

    unowned: list[dict[str, Any]] = []
    owned: dict[str, dict[str, Any]] = {}

    bar: Iterable[Any]
    with click.progressbar(files) as bar:
        for info in bar:
            resource = model.Session.scalar(
                sa.select(model.Resource).where(
                    model.Resource.id == "".join(info["location"].split("/")),
                ),
            )

            if resource:
                owned[resource.id] = info

            else:
                unowned.append(info)

    click.echo(f"Located owners for {len(owned)} files out of {content['count']}.")
    if click.confirm("Show user IDs and corresponding file?"):
        for key, info in owned.items():
            click.echo(f"{key}: {info['location']}")

    if unowned and click.confirm("Show files that do not belong to any user?"):
        for info in unowned:
            click.echo(f"{info['location']}")

    if click.confirm(
        "Transfer file ownership to resources identified in previous steps?",
    ):
        with click.progressbar(owned.items()) as bar:
            for resource_id, info in bar:
                bar.label = f"Transfering {info['location']}"

                tk.get_action("files_transfer_ownership")(
                    {"ignore_auth": True},
                    {
                        "pin": True,
                        "force": True,
                        "id": info["id"],
                        "owner_type": "resource",
                        "owner_id": resource_id,
                    },
                )
    packages: set[str] = set()
    if click.confirm("Modify resources, changing their url_type and url fields?"):
        with click.progressbar(owned.items()) as bar:
            for resource_id, info in bar:
                res = model.Resource.get(resource_id)
                fileobj = model.Session.get(shared.File, info["id"])
                if not fileobj:
                    tk.error_shout(
                        f"Filee {info['id']} disappeared during migration",
                    )
                    raise click.Abort

                if not res:
                    tk.error_shout(
                        f"Resource {resource_id} disappeared during migration",
                    )
                    raise click.Abort

                if not res.package_id:
                    tk.error_shout(f"Resource {resource_id} has no package_id")
                    raise click.Abort

                if res.url_type != "upload":
                    tk.error_shout(
                        f"Resource {resource_id} has suspicious "
                        + f"url_type: {res.url_type}",
                    )
                    raise click.Abort

                res.url_type = "file"
                fileobj.name = res.url
                res.url = info["id"]
                packages.add(res.package_id)

    click.echo("Commit DB changes...")
    model.Session.commit()

    click.echo("Reindex modified packages...")
    for pkg in packages:
        rebuild(pkg)



from __future__ import annotations

from typing import Any

import pytest

import ckan.model as model
from ckan.tests.helpers import call_action

from ckanext.files.model import File

call_action: Any


@pytest.fixture()
def random_file(create_with_upload: Any, faker: Any):
    return create_with_upload(
        faker.binary(10),
        faker.file_name(),
        action="files_file_create",
        name=faker.name(),
    )


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_redis")
class TestFileCreate:
    def test_basic_file(self, create_with_upload: Any):
        filename = "file.txt"
        result = create_with_upload(
            "hello",
            filename,
            action="files_file_create",
            name="Test file.txt",
        )

        assert result["name"] == "Test_file.txt"


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_redis")
class TestFileDelete:
    def test_basic_delete(self, random_file: dict[str, Any]):
        q = model.Session.query(File)
        assert q.count() == 1
        call_action("files_file_delete", id=random_file["id"])
        assert not q.count()


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileShow:
    def test_basic_show(self, random_file: dict[str, Any]):
        result = call_action("files_file_show", id=random_file["id"])
        assert result["id"] == random_file["id"]

        result = call_action("files_file_show", id=random_file["id"])
        assert result["id"] == random_file["id"]


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_redis")
class TestTransferOwnership:
    def test_transfer_to_different_entity(
        self,
        random_file: dict[str, Any],
        user: dict[str, Any],
        package: dict[str, Any],
    ):
        call_action(
            "files_transfer_ownership",
            id=random_file["id"],
            owner_type="user",
            owner_id=user["id"],
        )
        file = model.Session.get(File, random_file["id"])
        assert file
        assert file.owner
        assert file.owner.id == user["id"]

        call_action(
            "files_transfer_ownership",
            id=random_file["id"],
            owner_type="package",
            owner_id=package["id"],
        )
        file = model.Session.get(File, random_file["id"])
        assert file
        assert file.owner
        assert file.owner.id == package["id"]



from __future__ import annotations

import pydoc
import textwrap

import click

from ckanext.files import base

from . import dev, file, maintain, migrate, stats, storage

__all__ = [
    "files",
]


@click.group(short_help="ckanext-files CLI commands")
def files():
    pass


files.add_command(dev.group, "dev")
files.add_command(migrate.group, "migrate")
files.add_command(stats.group, "stats")
files.add_command(storage.group, "storage")
files.add_command(maintain.group, "maintain")
files.add_command(file.group, "file")


@files.command()
@click.option("-v", "--verbose", is_flag=True, help="Show adapter's documentation")
@click.option("-H", "--include-hidden", is_flag=True, help="Show hidden adapters")
def adapters(verbose: bool, include_hidden: bool):
    """Show all awailable storage adapters."""
    for name in sorted(base.adapters):
        adapter = base.adapters[name]
        if adapter.hidden and not include_hidden:
            continue

        click.secho(
            f"{click.style(name, bold=True)} - {adapter.__module__}:{adapter.__name__}",
        )
        if verbose:
            doc = pydoc.getdoc(adapter)
            wrapped = textwrap.indent(doc, "\t")
            if wrapped:
                click.echo(wrapped)



from __future__ import annotations

from typing import Any, cast

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import authz, model
from ckan.types import AuthResult, Context

from ckanext.files import interfaces, shared, types, utils


def _owner_allows(
    context: Context,
    owner_type: str,
    owner_id: str,
    operation: types.OwnerOperation,
) -> bool:
    """Decide if user is allowed to perform operation on owner."""
    for plugin in p.PluginImplementations(interfaces.IFiles):
        result = plugin.files_owner_allows(context, owner_type, owner_id, operation)
        if result is not None:
            return result

    if (operation == "file_transfer" and shared.config.transfer_as_update()) or (
        operation == "file_scan" and shared.config.scan_as_update()
    ):
        func_name = f"{owner_type}_update"

    else:
        func_name = f"{owner_type}_{operation}"

    try:
        tk.check_access(
            func_name,
            tk.fresh_context(context),
            {"id": owner_id},
        )

    except (tk.NotAuthorized, ValueError):
        return False

    return True


def _file_allows(
    context: Context,
    file: shared.File | shared.Multipart,
    operation: types.FileOperation,
) -> bool:
    """Decide if user is allowed to perform operation on file."""
    for plugin in p.PluginImplementations(interfaces.IFiles):
        result = plugin.files_file_allows(context, file, operation)
        if result is not None:
            return result

    info = file.owner_info if file else None

    if not info:
        return False

    cascade = shared.config.cascade_access()
    if info.owner_type not in cascade:
        return False

    if cascade[info.owner_type] and file.storage not in cascade[info.owner_type]:
        return False

    func_name = f"{info.owner_type}_{operation}"

    try:
        tk.check_access(
            func_name,
            tk.fresh_context(context),
            {"id": info.owner_id},
        )

    except (tk.NotAuthorized, ValueError):
        return False

    return True


def _get_user(context: Context) -> model.User | None:
    if "auth_user_obj" in context:
        return cast(model.User, context["auth_user_obj"])

    user = tk.current_user if tk.current_user.is_authenticated else None
    username = context["user"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

    if user and username == user.name:
        return cast(model.User, user)

    cache = utils.ContextCache(context)
    return cache.get("user", username, lambda: model.User.get(username))


def _get_file(
    context: Context,
    file_id: str,
    completed: bool,
) -> shared.File | shared.Multipart | None:
    cache = utils.ContextCache(context)
    return cache.get_model(
        "file",
        file_id,
        shared.File if completed else shared.Multipart,
    )


@tk.auth_allow_anonymous_access
def files_manage_files(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return {"success": False}


@tk.auth_allow_anonymous_access
def files_owns_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    if authz.is_authorized_boolean("files_manage_files", context, data_dict):
        return {"success": True}

    user = _get_user(context)
    if not user:
        return {
            "success": False,
            "msg": "Not an owner of the file",
        }

    file = _get_file(context, data_dict["id"], data_dict.get("completed", True))
    if not file or not file.owner_info:
        return {
            "success": False,
            "msg": "Not an owner of the file",
        }

    return {
        "success": file.owner_info.owner_type == "user"
        and file.owner_info.owner_id == user.id,
        "msg": "Not an owner of the file",
    }


@tk.auth_allow_anonymous_access
def files_edit_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    result = authz.is_authorized_boolean("files_owns_file", context, data_dict)
    if not result:
        file = _get_file(context, data_dict["id"], data_dict.get("completed", True))
        result = bool(file and _file_allows(context, file, "update"))

    return {"success": result, "msg": "Not allowed to edit file"}


@tk.auth_allow_anonymous_access
def files_read_file(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    result = authz.is_authorized_boolean("files_owns_file", context, data_dict)
    if not result:
        file = _get_file(context, data_dict["id"], data_dict.get("completed", True))
        result = bool(file and _file_allows(context, file, "show"))

    return {"success": result, "msg": "Not allowed to read file"}


@tk.auth_allow_anonymous_access
def files_file_search(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only file manager can search files."""
    return authz.is_authorized("files_manage_files", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_create(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    if context["user"] and (  # pyright: ignore[reportTypedDictNotRequiredAccess]
        shared.config.authenticated_uploads()
        and data_dict["storage"] in shared.config.authenticated_storages()
    ):
        return {"success": True}

    return authz.is_authorized("files_manage_files", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_delete(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can remove files."""
    result = authz.is_authorized_boolean("files_owns_file", context, data_dict)
    if not result:
        file = _get_file(context, data_dict["id"], data_dict.get("completed", True))
        result = bool(file and _file_allows(context, file, "delete"))

    return {"success": result, "msg": "Not allowed to delete file"}


@tk.auth_allow_anonymous_access
def files_file_replace(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can replace files."""
    return authz.is_authorized("files_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_show(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can view files."""
    return authz.is_authorized("files_read_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_download(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can download files."""
    return authz.is_authorized("files_read_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_file_rename(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only owner can rename files."""
    return authz.is_authorized("files_edit_file", context, data_dict)


@tk.auth_allow_anonymous_access
def files_multipart_refresh(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized(
        "files_edit_file",
        context,
        dict(data_dict, completed=False),
    )


@tk.auth_allow_anonymous_access
def files_multipart_start(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized(
        "files_file_create",
        context,
        dict(data_dict, completed=False),
    )


@tk.auth_allow_anonymous_access
def files_multipart_update(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized(
        "files_edit_file",
        context,
        dict(data_dict, completed=False),
    )


@tk.auth_allow_anonymous_access
def files_multipart_complete(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized(
        "files_edit_file",
        context,
        dict(data_dict, completed=False),
    )


@tk.auth_allow_anonymous_access
def files_file_pin(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized(
        "files_edit_file",
        context,
        dict(data_dict, completed=False),
    )


@tk.auth_allow_anonymous_access
def files_file_unpin(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    return authz.is_authorized(
        "files_edit_file",
        context,
        dict(data_dict, completed=False),
    )


@tk.auth_allow_anonymous_access
def files_transfer_ownership(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only file manager can transfer ownership."""
    file = _get_file(context, data_dict["id"], data_dict.get("completed", True))
    if file and file.owner_info and file.owner_info.pinned and not data_dict["force"]:
        return {"success": False, "msg": "File is pinned"}

    result = authz.is_authorized_boolean("files_manage_files", context, data_dict)
    if not result:
        result = bool(
            file
            and _file_allows(context, file, "update")
            and _owner_allows(
                context,
                data_dict["owner_type"],
                data_dict["owner_id"],
                "file_transfer",
            ),
        )

    return {"success": result, "msg": "Not allowed to edit file"}


@tk.auth_allow_anonymous_access
def files_file_scan(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Only file owner can list files."""
    result = authz.is_authorized_boolean("files_manage_files", context, data_dict)
    if not result:
        result = _owner_allows(
            context,
            data_dict["owner_type"],
            data_dict["owner_id"],
            "file_scan",
        )

    return {"success": result, "msg": "Not allowed to list files"}


@tk.auth_allow_anonymous_access
def files_resource_upload(context: Context, data_dict: dict[str, Any]) -> AuthResult:
    """Users can upload resources files."""
    # TODO: add restriction on number of free-files or their total size inside
    # resource storage.
    try:
        shared.get_storage(shared.config.resources_storage())
    except shared.exc.UnknownStorageError:
        return {"success": False}

    if data_dict.get("resource_id"):
        return authz.is_authorized(
            "resource_update", context, {"id": data_dict["resource_id"]}
        )

    return authz.is_authorized("resource_create", context, data_dict)


@tk.auth_allow_anonymous_access
def files_autocomplete_available_resource_files(
    context: Context,
    data_dict: dict[str, Any],
) -> AuthResult:
    """Users allowed to search their free resource files."""
    return {"success": True}



"""single owner.

Revision ID: a11926fa799e
Revises: c7081d7f02e8
Create Date: 2024-06-09 18:18:13.987236

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a11926fa799e"
down_revision = "c7081d7f02e8"
branch_labels = None
depends_on = None

table = sa.table("files_owner", sa.column("owner_type"))


def upgrade():
    op.drop_column("files_owner", "access")
    op.drop_column("files_owner", "id")
    op.add_column(
        "files_owner",
        sa.Column("pinned", sa.Boolean, nullable=False, server_default="false"),
    )

    op.create_primary_key("files_owner_pkey", "files_owner", ["item_id", "item_type"])
    op.drop_constraint(
        "files_owner_item_id_item_type_owner_id_owner_type_key",
        "files_owner",
    )


def downgrade():
    op.drop_column("files_owner", "pinned")
    op.add_column(
        "files_owner",
        sa.Column("access", sa.Text, nullable=False, server_default="full"),
    )
    op.drop_constraint("files_owner_pkey", "files_owner")
    op.add_column(
        "files_owner",
        sa.Column(
            "id",
            sa.Text(),
            primary_key=True,
            server_default=sa.text("uuid_in(md5(random()::text)::cstring)"),
        ),
    )
    op.create_primary_key("files_owner_pkey", "files_owner", ["id"])
    op.alter_column("files_owner", "id", server_default=None)

    op.create_unique_constraint(
        "files_owner_item_id_item_type_owner_id_owner_type_key",
        "files_owner",
        [
            "item_id",
            "item_type",
            "owner_id",
            "owner_type",
        ],
    )



"""Internal utilities of the extension.

Do not use this module outside of the extension and do not import any other
internal module except for config, types and exceptions. Only independent tools
are stored here, to avoid import cycles.

"""

from __future__ import annotations

import logging
from typing import Any, Callable, Iterable, TypeVar, cast

import file_keeper as fk
import jwt
from sqlalchemy.orm import Mapper

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.api_token import _get_algorithm, _get_secret  # type: ignore

from ckanext.files import types

log = logging.getLogger(__name__)

T = TypeVar("T")

SAMPLE_SIZE = 1024 * 2


owner_getters = fk.Registry[Callable[[str], Any]]({})


def is_supported_type(content_type: str, supported: Iterable[str]) -> bool:
    """Check whether content_type it matches supported types."""
    maintype, subtype = content_type.split("/")
    desired = {content_type, maintype, subtype}
    return any(st in desired for st in supported)


def get_owner(owner_type: str, owner_id: str):
    if getter := owner_getters.get(owner_type):
        return getter(owner_id)

    owner_model = "group" if owner_type == "organization" else owner_type
    mappers: Iterable[Mapper]
    if tk.check_ckan_version("2.11"):
        mappers = model.registry.mappers
    else:
        mappers = cast(
            "Iterable[Mapper]",
            tk.BaseModel._sa_registry.mappers
            | model.User._sa_class_manager.registry.mappers,  # type: ignore
        )

    for mapper in mappers:
        cls = mapper.class_
        table = getattr(cls, "__table__", None)
        if table is None:
            table = getattr(mapper, "local_table", None)

        if table is not None and table.name == owner_model:
            return model.Session.get(cls, owner_id)

    log.warning("Unknown owner type %s with ID %s", owner_type, owner_id)
    return None


def encode_token(data: dict[str, Any]) -> str:
    return jwt.encode(data, _get_secret(encode=True), algorithm=_get_algorithm())


def decode_token(token: str) -> dict[str, Any]:
    return jwt.decode(token, _get_secret(encode=False), algorithms=[_get_algorithm()])


class ContextCache:
    key = "files_cache"

    def __init__(self, context: types.Context):
        self.session = context.get("session", model.Session)
        self.cache = cast("dict[str, Any]", context).setdefault(self.key, {})

    def invalidate(self, type: str, id: str):
        self.cache.setdefault(type, {}).pop(id, None)

    def bucket(self, type: str) -> dict[str, Any]:
        return self.cache.setdefault(type, {})

    def set(self, type: str, id: str, value: T) -> T:
        bucket = self.bucket(type)
        bucket[id] = value
        return value

    def get(self, type: str, id: str, compute: Callable[[], T]) -> T:
        bucket = self.bucket(type)
        if id not in bucket:
            bucket[id] = compute()

        return bucket[id]

    def get_model(self, type: str, id: str, model_class: type[T]) -> T | None:
        return self.get(type, id, lambda: self.session.get(model_class, id))



from __future__ import annotations

import json
import os
from typing import Any

import file_keeper as fk
import yaml

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model
from ckan.exceptions import CkanConfigurationException
from ckan.logic import clear_validators_cache

from . import base, config, interfaces, shared, storage, types, utils


@tk.blanket.helpers
@tk.blanket.validators
@tk.blanket.actions
@tk.blanket.auth_functions
@tk.blanket.cli
@tk.blanket.blueprints
class FilesPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurable)
    p.implements(p.IConfigurer, inherit=True)
    p.implements(interfaces.IFiles, inherit=True)

    p.implements(p.IConfigDeclaration)

    def declare_config_options(self, declaration: types.Declaration, key: types.Key):
        # this call allows using custom validators in config declarations
        clear_validators_cache()

        here = os.path.dirname(__file__)
        with open(os.path.join(here, "config_declaration.yaml"), "rb") as src:
            declaration.load_dict(yaml.safe_load(src))

        _register_adapters()

        # add config declarations for configured storages. In this way user can
        # print all available options for every storage via `ckan config
        # declaration files`
        for name, settings in config.storages().items():
            # make base key so that storage can declare options by extending it
            # like `storage_key.option_name`, instead of logner form
            # `key.ckanext.files.storage.STORAGE_NAME.option_nam`
            storage_key = key.from_string(config.STORAGE_PREFIX + name)

            if tk.check_ckan_version("2.10.3"):
                # literal evaluation in validators was added in
                # v2.10.3. Without it we cannot validate dynamically storage
                # type
                available_adapters = json.dumps(
                    list(base.adapters),
                    separators=(",", ":"),
                )

                declaration.declare(
                    storage_key.type,
                    settings.get("type"),
                ).append_validators(
                    f"one_of({available_adapters})",
                ).set_description(
                    "Storage adapter used by the storage",
                )

            adapter = base.adapters.get(settings.get("type", ""))
            if not adapter:
                continue

            if func := getattr(adapter, "declare_config_options", None):
                func(declaration, storage_key)

    # IFiles
    def files_get_storage_adapters(self) -> dict[str, type[base.Storage]]:
        adapters: dict[str, type[base.Storage]] = {
            "files:fs": storage.FsStorage,
            "files:public_fs": storage.PublicFsStorage,
            "files:ckan_resource_fs": storage.CkanResourceFsStorage,
            "files:redis": storage.RedisStorage,
            "files:filebin": storage.FilebinStorage,
            "files:db": storage.DbStorage,
            "files:link": storage.LinkStorage,
        }

        if hasattr(storage, "S3Storage"):
            adapters.update({"files:s3": storage.S3Storage})

        if hasattr(storage, "GoogleCloudStorage"):
            adapters.update({"files:google_cloud_storage": storage.GoogleCloudStorage})

        if hasattr(storage, "OpenDalStorage"):
            adapters.update({"files:opendal": storage.OpenDalStorage})

        if hasattr(storage, "LibCloudStorage"):
            adapters.update({"files:libcloud": storage.LibCloudStorage})

        return adapters

    # IConfigurable
    def configure(self, config_: Any):
        _initialize_storages()
        _register_owner_getters()

    # IConfigurer
    def update_config(self, config_: Any):
        tk.add_template_directory(config_, "templates")
        tk.add_resource("assets", "files")
        tk.add_public_directory(config_, "public")

        # this template folder contains a single unsafe template used in
        # resource migration workflow described in README
        if config.override_resource_form():
            tk.add_template_directory(config_, "resource_form_templates")


def _register_adapters():
    """Register all storage types provided by extensions."""
    base.adapters.reset()
    fk.ext.register()

    for plugin in p.PluginImplementations(interfaces.IFiles):
        for name, adapter in plugin.files_get_storage_adapters().items():
            base.adapters.register(name, adapter)


def _initialize_storages():
    """Initialize all configured storages.

    Raise an exception if storage type is not registered.
    """
    base.storages.reset()
    for name, settings in config.storages().items():
        try:
            storage = base.make_storage(name, settings)
        except (
            shared.exc.UnknownAdapterError,
            shared.exc.InvalidStorageConfigurationError,
        ) as err:
            raise CkanConfigurationException(str(err)) from err

        base.storages.register(name, storage)


def _register_owner_getters():
    """Register functions used by Owner model to locate owner entity."""
    utils.owner_getters.reset()

    utils.owner_getters.register("user", model.User.get)
    utils.owner_getters.register("package", model.Package.get)
    utils.owner_getters.register("resource", model.Resource.get)
    utils.owner_getters.register("group", model.Group.get)
    utils.owner_getters.register("organization", model.Group.get)

    for plugin in p.PluginImplementations(interfaces.IFiles):
        for name, getter in plugin.files_register_owner_getters().items():
            utils.owner_getters.register(name, getter)



"""Interfaces of the extension."""

from __future__ import annotations

from typing import Any, Callable

from ckan.plugins import Interface

from ckanext.files import types

File = Multipart = Any


# --8<-- [start:interface]
class IFiles(Interface):
    """Extension point for ckanext-files.

    This interface is not stabilized. Implement it with `inherit=True`.

    Example:
        ```python
        class MyPlugin(p.SingletonPlugin):
            p.implements(interfaces.IFiles, inherit=True)
        ```
    """

    def files_get_storage_adapters(self) -> dict[str, Any]:
        """Return mapping of storage type to adapter class.

        Returns:
            adapters provided by the implementation

        Example:
            ```python
            def files_get_storage_adapters(self):
                return {
                    "my_ext:dropbox": DropboxStorage,
                }
            ```
        """
        return {}

    def files_register_owner_getters(self) -> dict[str, Callable[[str], Any]]:
        """Return mapping with lookup functions for owner types.

        Name of the getter is the name used as `Owner.owner_type`. The getter
        itself is a function that accepts owner ID and returns optional owner
        entity.

        Returns:
            getters for specific owner types

        Example:
            ```python
            def files_register_owner_getters(self):
                return {"resource": model.Resource.get}
            ```
        """
        return {}

    def files_file_allows(
        self,
        context: types.Context,
        file: File | Multipart,
        operation: types.FileOperation,
    ) -> bool | None:
        """Decide if user is allowed to perform specified operation on the file.

        Return True/False if user allowed/not allowed. Return `None` to rely on
        other plugins.

        Default implementation relies on cascade_access config option. If owner
        of file is included into cascade access, user can perform operation on
        file if he can perform the same operation with file's owner.

        If current owner is not affected by cascade access, user can perform
        operation on file only if user owns the file.

        Args:
            context: API context
            file: accessed file object
            operation: performed operation

        Returns:
            decision whether operation is allowed for the file

        Example:
            ```python
            def files_file_allows(
                    self, context,
                    file: shared.File | shared.Multipart,
                    operation: shared.types.FileOperation
            ) -> bool | None:
                if file.owner_info and file.owner_info.owner_type == "resource":
                    return is_authorized_boolean(
                        f"resource_{operation}",
                        context,
                        {"id": file.owner_info.id}
                    )

                return None
            ```
        """
        return None

    def files_owner_allows(
        self,
        context: types.Context,
        owner_type: str,
        owner_id: str,
        operation: types.OwnerOperation,
    ) -> bool | None:
        """Decide if user is allowed to perform specified operation on the owner.

        Return True/False if user allowed/not allowed. Return `None` to rely on
        other plugins.

        Args:
            context: API context
            owner_type: type of the tested owner
            owner_id: type of the tested owner
            operation: performed operation

        Returns:
            decision whether operation is allowed for the owner

        Example:
            ```python
            def files_owner_allows(
                    self, context,
                    owner_type: str, owner_id: str,
                    operation: shared.types.OwnerOperation
            ) -> bool | None:
                if owner_type == "resource" and operation == "file_transfer":
                    return is_authorized_boolean(
                        f"resource_update",
                        context,
                        {"id": owner_id}
                    )

                return None
            ```
        """
        return None


# --8<-- [start:interface]



"""alter files table.

Revision ID: d94bacc491e4
Revises: 76fdef67f479
Create Date: 2024-05-29 22:19:22.535787

"""

import sqlalchemy as sa
from alembic import context, op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "d94bacc491e4"
down_revision = "76fdef67f479"
branch_labels = None
depends_on = None

table = sa.table(
    "files_file",
    sa.column("id"),
    sa.column("name"),
    sa.column("storage"),
    sa.column("storage_data", JSONB),
    sa.column("plugin_data", JSONB),
    sa.column("location"),
    sa.column("content_type"),
    sa.column("size"),
    sa.column("ctime"),
    sa.column("hash"),
    sa.column("completed"),
)
owner_table = sa.table(
    "files_owner",
    sa.column("item_id"),
    sa.column("item_type"),
)


def upgrade():
    bind = op.get_bind()

    multipart_table = op.create_table(
        "files_multipart",
        sa.Column("id", sa.UnicodeText, primary_key=True),
        sa.Column("name", sa.UnicodeText, nullable=False),
        sa.Column("location", sa.UnicodeText, nullable=False, server_default=""),
        sa.Column("storage", sa.Text, nullable=False),
        sa.Column("ctime", sa.DateTime, server_default=sa.func.now()),
        sa.Column("size", sa.Integer, nullable=False, server_default="0"),
        sa.Column("content_type", sa.Text, nullable=False, server_default=""),
        sa.Column("hash", sa.Text, nullable=False, server_default=""),
        sa.Column("storage_data", JSONB, server_default="{}"),
        sa.Column("plugin_data", JSONB, server_default="{}"),
    )

    stmt = sa.select(
        table.c.id,
        table.c.name,
        table.c.storage,
        table.c.ctime,
        table.c.storage_data,
        table.c.plugin_data,
    ).where(table.c.completed == sa.false())

    if context.is_offline_mode():
        context.execute("""-- MANUAL MODE: start""")
        context.execute("""-- * select all completed=false files""")
        context.execute("""-- * move them into multipart table""")
        context.execute("""-- * save storage_data.filename as location""")
        context.execute("""-- * add default size=0""")
        context.execute("""-- * add default hash=''""")
        context.execute("""-- * change item type to multipart in owner table""")
        context.execute("""-- MANUAL MODE: end""")
    else:
        for id, name, storage, ctime, data, plugin_data in bind.execute(stmt):
            data["location"] = data.pop("filename")
            content_type = data.pop("content_type", "application/octet-stream")
            size = data.pop("size", 0)
            hash = data.pop("hash", "")

            bind.execute(
                sa.insert(multipart_table).values(
                    id=id,
                    name=name,
                    storage=storage,
                    ctime=ctime,
                    content_type=content_type,
                    size=size,
                    hash=hash,
                    storage_data=data,
                    plugin_data=plugin_data,
                ),
            )
            bind.execute(
                sa.update(owner_table)
                .values(item_type="multipart")
                .where(owner_table.c.item_id == id),
            )

        bind.execute(sa.delete(table).where(table.c.completed == sa.false()))

    op.add_column("files_file", sa.Column("location", sa.Text))
    op.add_column(
        "files_file",
        sa.Column("content_type", sa.Text, server_default="application/octet-stream"),
    )
    op.add_column("files_file", sa.Column("size", sa.Integer, server_default="0"))
    op.add_column("files_file", sa.Column("hash", sa.Text, server_default=""))
    op.drop_column("files_file", "completed")

    stmt = sa.select(table.c.id, table.c.storage_data)
    if context.is_offline_mode():
        context.execute("""-- MANUAL MODE: start""")
        context.execute("""-- * extract storage_data.filename into location""")
        context.execute(
            """-- * extract content_type, size and hash from storage data to top"""
        )
        context.execute("""-- MANUAL MODE: end""")
    else:
        for id, data in bind.execute(stmt):
            location = data.pop("filename")
            content_type = data.pop("content_type")
            size = data.pop("size")
            hash = data.pop("hash")

            bind.execute(
                sa.update(table)
                .values(
                    location=location,
                    storage_data=data,
                    content_type=content_type,
                    size=size,
                    hash=hash,
                )
                .where(table.c.id == id),
            )

    op.alter_column("files_file", "location", nullable=False)


def downgrade():
    bind = op.get_bind()

    stmt = sa.select(
        table.c.id,
        table.c.storage_data,
        table.c.location,
        table.c.content_type,
        table.c.size,
        table.c.hash,
    )
    if context.is_offline_mode():
        context.execute("""-- MANUAL MODE: start""")
        context.execute("""-- * move content_type, size, hash into storage_data""")
        context.execute("""-- * move location into storage_data.filename""")
        context.execute("""-- MANUAL MODE: end""")
    else:
        for id, data, location, content_type, size, hash in bind.execute(stmt):
            data["filename"] = location
            data["content_type"] = content_type
            data["size"] = size
            data["hash"] = hash
            bind.execute(
                sa.update(table).values(storage_data=data).where(table.c.id == id)
            )

    op.drop_column("files_file", "location")
    op.drop_column("files_file", "content_type")
    op.drop_column("files_file", "size")
    op.drop_column("files_file", "hash")
    op.add_column(
        "files_file",
        sa.Column("completed", sa.Boolean, server_default="false"),
    )

    stmt = sa.update(table).values(completed=sa.true())
    if context.is_offline_mode():
        context.execute(stmt)
    else:
        bind.execute(stmt)

    op.drop_table("files_multipart")



from __future__ import annotations

import dataclasses

from file_keeper.default.adapters import sqlalchemy

from ckanext.files import shared, types


@dataclasses.dataclass()
class Settings(shared.Settings, sqlalchemy.Settings):
    pass


class DbStorage(shared.Storage, sqlalchemy.SqlAlchemyStorage):
    hidden = True

    settings: Settings  # type: ignore
    SettingsFactory = Settings

    @classmethod
    def declare_config_options(
        cls,
        declaration: types.Declaration,
        key: types.Key,
    ):
        declaration.declare(key.db_url).required()
        declaration.declare(key.table_name).required()
        declaration.declare(key.location_column).required()
        declaration.declare(key.content_column).required()



from __future__ import annotations

import base64
import dataclasses

from file_keeper.default.adapters import s3

from ckan.config.declaration import Declaration, Key

from ckanext.files import shared


def decode(value: str) -> str:
    return base64.decodebytes(value.encode()).hex()


@dataclasses.dataclass()
class Settings(shared.Settings, s3.Settings):
    pass


class S3Storage(shared.Storage, s3.S3Storage):
    hidden = True
    settings: Settings  # type: ignore
    SettingsFactory = Settings

    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)

        declaration.declare(key.bucket).required().set_description(
            "Name of the S3 bucket where uploaded data will be stored.",
        )
        declaration.declare(key.path, "").set_description(
            "Path to the folder where uploaded data will be stored.",
        )
        declaration.declare(key.key).set_description("The access key for AWS account.")
        declaration.declare(key.secret).set_description(
            "The secret key for AWS account."
        )

        declaration.declare(key.region).set_description(
            "The AWS Region used in instantiating the client."
        )

        declaration.declare(key.endpoint).set_description(
            "The complete URL to use for the constructed client."
        )



"""Task queue logic.

Create a context for scheduling tasks upon exiting the context.

Initially was designed for `with_task_queue` decorator. It creates task queue
upon entering the API action and keeps track of ownership transfer tasks
produced by validators. If all validators succesfully passed and decorated
action finished, tasks are executed, moving file ownership to the
created/updated entity.

It's a high-level module that relies on base and utils.

"""

from __future__ import annotations

import abc
import contextvars
import dataclasses
import functools
from collections import deque
from typing import Any, Literal

import file_keeper as fk

import ckan.plugins.toolkit as tk
from ckan.types import FlattenKey

from ckanext.files import base, types

_task_queue: contextvars.ContextVar[deque[types.PTask] | None] = contextvars.ContextVar(
    "transfer_queue",
    default=None,
)


class TaskQueue:
    """Thread-safe context for managing tasks.

    Example:
        ```python
        queue = TaskQueue()
        with queue:
            function_that_adds_tasks_to_queue()
        data_passed_into_tasks = ...
        queue.process(data_passed_into_tasks)
        ```
    """

    # container for tasks
    queue: deque[types.PTask]

    # refresh token that restores previous state of queue
    token: contextvars.Token[Any] | None

    def __len__(self):
        return len(self.queue)

    def __init__(self):
        self.queue = deque()
        self.token = None

    def __enter__(self):
        # save token to restore previous queue. Required for nested queues
        self.token = _task_queue.set(self.queue)
        return self.queue

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        # restore previous queue state. Required for nested queues
        if self.token:
            _task_queue.reset(self.token)

    def process(self, data: dict[str, Any]):
        """Execute queued tasks in FIFO order.

        `data` is passed to every task as a first argument. In addition, task
        receives its position in execution queue and results of the previous
        task. The first task receives Task.NO_PREVIOUS_TASK as third argument,
        because there are no results from previous task yet.
        """
        prev = Task.NO_PREVIOUS_TASK
        idx = 0
        while self.queue:
            task = self.queue.popleft()
            prev = task(data, idx, prev)
            idx += 1


class Task(abc.ABC):
    """Base task for TaskQueue.

    The only requirement for subclasses is implementing Task.run.
    """

    NO_PREVIOUS_TASK = object()

    @staticmethod
    def extract(source: dict[str, Any], path: FlattenKey):
        """Extract value from dictionary using FlattenKey from validators.

        Args:
            source: dictionary with data
            path: path to the extracted member from source

        Example:
            ```python
            data = {"a": {"b": {"c": 42}}}
            assert Task.extract(data, ("a", "b", "c")) == 42
            ```
        """
        for step in path:
            source = source[step]

        return source

    def __call__(self, result: Any, idx: int, prev: Any):
        return self.run(result, idx, prev)

    @abc.abstractmethod
    def run(self, result: Any, idx: int, prev: Any):
        """Execute task.

        `result` is an arbitrary data passed into every task in the
        queue. `idx` reflects current task's position in queue. `prev` contains
        result of previous task or Task.NO_PREVIOUS_TASK if current task is the
        first in the queue.
        """
        ...


@dataclasses.dataclass
class OwnershipTransferTask(Task):
    """Taks for transfering ownership to not-yet-existing entity.

    Designed for scheduling from validators:

    Example:
        ```python
        def validator(key, data, errors, context):
            file_id = data[key]
            owner_id_path = key[:-1] + ("id",)
            task = OwnershipTransferTask(file_id, "resource", owner_id_path)
            add_task(task)
        ```

    This task requires the ID of the file to trasfer, type of the future owner
    and flattened path to ID field of the owner.

    Example:
        ```python
        task = OwnershipTransferTask('xxx', 'resource', ("resources", 0, "id"))
        dataset = {"name": "test", "resources": [{"id": "123"}]}
        task.run(dataset, 0, Task.NO_PREVIOUS_TASK)
        ```
    """

    # ID of the transfered file
    file_id: str
    # type of the future owner
    owner_type: str
    # flattened path to ID of future owner in execution data
    destination: FlattenKey | str

    strategy: str = "path"

    def run(self, result: dict[str, Any], idx: int, prev: Any):
        """Transfer file ownership.

        Ownerhip transfered to an entity whose ID is stored in `result` under
        `self.id_path` path.
        """
        if self.strategy == "path" and isinstance(self.destination, tuple):
            owner_id = self.extract(result, self.destination)

        elif self.strategy == "value":
            owner_id = self.destination

        else:
            raise ValueError((self.strategy, self.destination))

        return tk.get_action("files_transfer_ownership")(
            {"ignore_auth": True},
            {
                "id": self.file_id,
                "owner_type": self.owner_type,
                "owner_id": owner_id,
                # "force": True,
                "pin": True,
            },
        )


@dataclasses.dataclass
class UploadAndAttachTask(Task):
    """Taks for creating file and transfering it to specified owner.

    Optionally, task can update owner using specified API action.

    Designed for scheduling from validators:
    >>> def validator(key, data, errors, context):
    >>>     upload = data[key]
    >>>     owner_id_path = key[:-1] + ("id",)
    >>>     task = UploadAndAttachTask("default, upload, "resource", owner_id_path)
    >>>     add_task(task)

    This task requires the name of the storage for uploaded file, upload
    object, type of the future owner and flattened path to ID field of the
    owner.

    Optionally, it accepts action name for patching owner entity, name of the
    file's property that will be copied into owner entity and field of owner
    entity to copy file's property into it.

    Example:
    >>> task = UploadAndAttachTask(
    >>>     "default",
    >>>     make_upload(b"file.txt", "hello world),
    >>>     "resource",
    >>>     ("resources", 0, "id"),
    >>>     "id",
    >>>     "resource_patch",
    >>>     "attachment_id",
    >>> )
    >>> dataset = {"name": "test", "resources": [{"id": "123", "attachment_id": None}]}
    >>> task.run(dataset, 0, Task.NO_PREVIOUS_TASK)

    """

    # storage for file upload
    storage: str
    # upload object with file's content
    upload: fk.Upload
    # type of future owner
    owner_type: str
    # flattened path to owner's ID field
    destination: FlattenKey | str

    # property of the file that will be added to owner data. Currently
    # supported only id and public_url.
    attach_as: Literal["id", "permanent_url"] | None
    # action to use for patching owner's data
    action: str | None = None
    # field in owner's data that will hold property of the current file
    destination_field: str | None = None

    strategy: str = "path"

    def run(self, result: dict[str, Any], idx: int, prev: Any):
        """Upload file, transfer ownership and, optionally, patch the owner."""
        info = tk.get_action("files_file_create")(
            {"ignore_auth": True},
            {"upload": self.upload, "storage": self.storage},
        )

        if self.strategy == "path" and isinstance(self.destination, tuple):
            owner_id = self.extract(result, self.destination)

        elif self.strategy == "value":
            owner_id = self.destination

        else:
            raise ValueError((self.strategy, self.destination))

        info = tk.get_action("files_transfer_ownership")(
            {"ignore_auth": True},
            {
                "id": info["id"],
                "owner_type": self.owner_type,
                "owner_id": owner_id,
                # "force": True,
                "pin": True,
            },
        )

        if self.attach_as and self.action and self.destination_field:
            if self.attach_as == "permanent_url":
                storage = base.get_storage(self.storage)
                value = storage.permanent_link(base.FileData.from_dict(info))
            else:
                value = info["id"]

            user = tk.get_action("get_site_user")({"ignore_auth": True}, {})
            tk.get_action(self.action)(
                {"ignore_auth": True, "user": user["name"]},
                {"id": info["owner_id"], self.destination_field: value},
            )

        return info


def with_task_queue(func: Any, name: str | None = None):
    """Decorator for functions that schedule tasks.

    Decorated function automatically initializes separate task queue which is
    processed when function is finished. All tasks receive function's result as
    execution data(first argument of `Task.run`).

    Without this decorator, you have to manually create task queue context
    before queuing tasks.

    Example:
        ```python
        @with_task_queue
        def my_action(context, data_dict):
            ...
        ```
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        queue = TaskQueue()
        with queue:
            result = func(*args, *kwargs)
            queue.process(result)

        return result

    if name:
        wrapper.__name__ = name
    return wrapper


class OutOfQueueError(RuntimeError):
    """Attempt to add task without initializing task queue context."""

    def __str__(self):
        return "Task queue accessed outside of queue context"


def add_task(task: types.PTask):
    """Add task to the current task queue.

    This function can be called only inside task queue context. Such context
    initialized automatically inside functions decorated with
    `with_task_queue`:

    Example:
        ```python
        @with_task_queue
        def taks_producer():
            add_task(...)

        task_producer()
        ```

    Task queue context can be initialized manually using TaskQueue and
    `with` statement:

    Example:
        ```python
        queue = TaskQueue()
        with queue:
            add_task(...)

        queue.process(execution_data)
        ```
    """
    queue = _task_queue.get()
    if queue is None:
        raise OutOfQueueError
    queue.append(task)



from __future__ import annotations

import os
from datetime import datetime
from typing import Any, Iterable, TypedDict

import ckan.plugins.toolkit as tk
from ckan import model

from . import shared

HERE = os.path.dirname(__file__)


class LinkDetails(TypedDict):
    label: str
    content_type: str
    size: int
    href: str | None


def files_parse_tz_date(value: str, strip_timezone: bool = False):
    """Transform timezone-aware ISO 8601 value into datetime object."""
    result = datetime.fromisoformat(value)
    if strip_timezone:
        result = result.replace(tzinfo=None)

    return result


def files_humanize_content_type(content_type: str) -> str:
    if not content_type:
        content_type = "application/octet-stream"

    # for name, types in tk.h.resource_formats().items():
    #     if name == content_type or content_type in types:
    #         return types[2]

    main, _sub = content_type.split("/")
    return main.capitalize()


def files_content_type_icon(
    content_type: str,
    theme: str,
    extension: str,
) -> str | None:
    if not content_type:
        content_type = "application/octet-stream"

    url = "ckanext-files/mimetype_icons/{}/{}.{}".format(
        theme,
        content_type.replace("/", "-"),
        extension,
    )

    if os.path.exists(os.path.join(HERE, "public", url)):
        return tk.h.url_for_static(url)

    return None


def files_get_file(file_id: str | None) -> shared.File | None:
    if not file_id:
        return None
    return model.Session.get(shared.File, file_id)


def files_link_details(
    file_id: str,
    *types: Iterable[str],
    **kwargs: Any,
) -> LinkDetails | None:
    file = model.Session.get(shared.File, file_id)

    if not file:
        return None

    storage = shared.get_storage(file.storage)

    for lt in types or ("permanent", "temporal"):
        func = getattr(storage, f"{lt}_link", None)
        if not func:
            continue
        if link := func(shared.FileData.from_object(file), **kwargs):
            return {
                "label": file.name,
                "content_type": file.content_type,
                "size": file.size,
                "href": link,
            }


def _is_storage_configured(name: str) -> bool:
    try:
        return bool(shared.get_storage(name))
    except shared.exc.UnknownStorageError:
        return False


def files_group_images_storage_is_configured() -> bool:
    return _is_storage_configured(shared.config.group_images_storage())


def files_user_images_storage_is_configured() -> bool:
    return _is_storage_configured(shared.config.user_images_storage())


def files_resources_storage_is_configured() -> bool:
    return _is_storage_configured(shared.config.resources_storage())



from __future__ import annotations

import logging
from typing import Any, Literal, cast

import file_keeper as fk

import ckan.plugins.toolkit as tk
from ckan import authz
from ckan.types import Context, FlattenDataDict, FlattenErrorDict, FlattenKey

from ckanext.files import shared, task, utils

log = logging.getLogger(__name__)


def files_cascade_options(value: Any) -> dict[str, set[str]]:
    if isinstance(value, str):
        value = value.split()

    if isinstance(value, list):
        cascade: dict[str, set[str]] = {}
        for item in cast("list[str]", value):
            type, *rest = item.split(":", 1)
            cascade.setdefault(type, set()).update(rest)
        value = cascade

    if isinstance(value, dict):
        return cast("dict[str, set[str]]", value)

    msg = "Cascade rules are not correct"
    raise tk.Invalid(msg)


# unstable
def files_skip_absolute_url(value: Any):
    """Stop validation and accept value if it's an absolute URL."""
    if isinstance(value, str) and value.startswith(("https://", "http://")):
        raise tk.StopOnError


# unstable
def files_verify_url_type_and_value(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Skip validation for non-file resources.

    Remove everything before last `/` for file resources.
    """
    if data.get(key[:-1] + ("url_type",)) != "file":
        raise tk.StopOnError

    value = data[key]
    if value:
        data[key] = value.rsplit("/", 1)[-1]


# unstable
def files_id_into_resource_download_url(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Transform file ID into resource's download URL."""
    package_id = data.get(key[:-1] + ("package_id",))
    resource_id = data.get(key[:-1] + ("id",))

    data[key] = tk.url_for(
        "files.resource_download",
        id=package_id,
        resource_id=resource_id,
        filename=data[key],
        _external=True,
    )


# unstable
def files_file_into_public_url(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Transform ID into public URL or raise error if not supported."""
    value: str | list[str] = data[key]
    use_list = isinstance(value, list)
    ids = value if use_list else [value]
    result = []
    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    for file_id in ids:
        file = sess.get(shared.File, file_id)
        if not file:
            msg = "File does not exist"
            errors[key].append(msg)
            raise tk.StopOnError

        info = shared.FileData.from_object(file)
        storage = shared.get_storage(file.storage)
        url = storage.permanent_link(info)
        if not url:
            msg = "File does not support permanent URLs"
            errors[key].append(msg)
            raise tk.StopOnError
        result.append(url)

    data[key] = result if use_list else result[0]


def files_into_upload(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Convert value into Upload object."""
    try:
        data[key] = fk.make_upload(data[key])

    except TypeError as err:
        msg = f"Unsupported source type: {err}"
        errors[key].append(msg)
        raise tk.StopOnError from err

    except ValueError as err:
        msg = f"Wrong file: {err}"
        errors[key].append(msg)
        raise tk.StopOnError from err


def files_parse_filesize(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Convert human-readable filesize into an integer."""
    value = data[key]
    if isinstance(value, int):
        return

    try:
        data[key] = fk.parse_filesize(value)
    except ValueError as err:
        msg = f"Wrong filesize string: {value}"
        errors[key].append(msg)
        raise tk.StopOnError from err


def files_ensure_name(name_field: str):
    """Apply to Upload to guess filename where `name_field` is empty."""

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        name_key = key[:-1] + (name_field,)
        if data.get(name_key):
            return

        if name := data[key].filename:
            data[name_key] = name
            return

        msg = f"Name is missing and cannot be deduced from {key[-1]}"
        errors[key].append(msg)
        raise tk.StopOnError

    return validator


def files_file_id_exists(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Verify that file ID exists."""
    value = data[key]
    use_list = isinstance(value, list)
    ids: str | list[str] = value if use_list else [value]

    sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
    for file_id in ids:
        file = sess.get(shared.File, file_id)
        if not file:
            msg = "File does not exist"
            errors[key].append(msg)
            raise tk.StopOnError


# unstable
def files_content_type_from_file(file_field: str, if_empty: bool = False):
    """Copy MIMEtype of the file in specified field."""

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        value = data[key]
        if value and if_empty:
            return

        file_id = data.get(key[:-1] + (file_field,))
        if not file_id:
            return
        file_id = file_id.rsplit("/", 1)[-1]
        sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        file = sess.get(shared.File, file_id)

        if file:
            data[key] = file.content_type

    return validator


def files_accept_file_with_type(*supported_types: str):
    """Verify that file has allowed MIMEtype."""

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        value: str | list[str] = data[key]

        ids = value if isinstance(value, list) else [value]
        sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        for file_id in ids:
            file = sess.get(shared.File, file_id)
            if not file:
                msg = "File does not exist"
                errors[key].append(msg)
                raise tk.StopOnError

            actual = file.content_type

            if not utils.is_supported_type(actual, supported_types):
                expected = ", ".join(supported_types)
                msg = (
                    f"Type {actual} is not supported."
                    + f" Use one of the following types: {expected}"
                )
                errors[key].append(msg)
                raise tk.StopOnError

    return validator


def files_accept_file_with_storage(*supported_storages: str):
    """Verify that file stored inside specified storage."""

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        value: str | list[str] = data[key]

        ids = value if isinstance(value, list) else [value]
        sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

        for file_id in ids:
            file = sess.get(shared.File, file_id)
            if not file:
                msg = "File does not exist"
                errors[key].append(msg)
                raise tk.StopOnError

            if file.storage not in supported_storages:
                expected = ", ".join(supported_storages)
                msg = (
                    f"Storage {file.storage} is not supported."
                    + f" Use one of the following storages: {expected}"
                )
                errors[key].append(msg)
                raise tk.StopOnError

    return validator


def files_transfer_ownership(owner_type: str, id_field: str | list[str] = "id"):  # noqa: C901
    """Tranfer file ownership to validated object."""
    if isinstance(id_field, str):
        id_field = [id_field]

    def validator(  # noqa: C901
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        msg = "Is not an owner of the file"

        value = data[key]
        id_field_path = key[:-1]
        for step in id_field:
            id_field_path = (
                id_field_path[:-1] if step == ".." else id_field_path + (step,)
            )

        if data.get(id_field_path):
            strategy = "value"
            task_destination = data[id_field_path]

        elif len(key) == 1:
            strategy = "path"
            task_destination = id_field_path

        else:
            max_idx = 0
            idx_position = len(key) - 2
            for flat_field in data:
                if flat_field[:idx_position] != key[:idx_position]:
                    continue
                max_idx = max(key[idx_position], flat_field[idx_position])
            strategy = "path"
            task_destination = key[:-2] + (key[-2] - max_idx - 1,)
            for step in id_field:
                task_destination = (
                    task_destination[:-1]
                    if step == ".."
                    else task_destination + (step,)
                )

        ids: list[str] = value if isinstance(value, list) else [value]
        user = authz._get_user(context.get("user"))  # type: ignore
        sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

        for file_id in ids:
            file = sess.get(shared.File, file_id)
            if not file or not file.owner_info:
                errors[key].append(msg)
                raise tk.StopOnError

            owner_id = data.get(id_field_path)
            actual = file.owner_info.owner_type, file.owner_info.owner_id

            if actual == (owner_type, owner_id):
                continue

            if not user or actual != ("user", user.id):
                errors[key].append(msg)
                continue

            if file.owner_info.pinned:
                errors[key].append("File is pinned")
                continue

            shared.add_task(
                task.OwnershipTransferTask(
                    file_id, owner_type, task_destination, strategy
                ),
            )

    return validator


# unstable
def files_update_resource_url(
    key: FlattenKey,
    data: FlattenDataDict,
    errors: FlattenErrorDict,
    context: Context,
):
    """Transform file ID into resource's download URL."""
    data[key[:-1] + ("url",)] = tk.url_for(
        "files.dispatch_download",
        file_id=data[key],
        _external=True,
    )


# unstable
def files_set_field(field: str, value: Any):
    """Change value of the field to specified value."""

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        data[key[:-1] + (field,)] = value

    return validator


# unstable
def files_copy_attribute(attribute: str, destination: str):
    """Copy file's attribute into a different field."""

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        value = data[key]
        sess = context["session"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

        file = sess.get(shared.File, value)
        if not file or not hasattr(file, attribute):
            return

        data[key[:-1] + (destination,)] = getattr(file, attribute)

    return validator


# unstable
def files_validate_with_storage(storage_name: str):
    """Apply storage validators to file."""
    storage = shared.get_storage(storage_name)

    def validator(value: shared.Upload):
        if isinstance(storage, shared.Storage):
            try:
                storage.validate_content_type(value.content_type)
                storage.validate_size(value.size)
            except shared.exc.UploadError as err:
                raise tk.Invalid(str(err)) from err
        return value

    return validator


# unstable
def files_upload_as(  # noqa: PLR0913
    storage: str,
    owner_type: str,
    id_field: str,
    attach_as: Literal["id", "permanent_url"] | None,
    using_action: str | None = None,
    destination_field: str | None = None,
):
    """This validator exists for backward compatibility and will be removed.

    Schedule file upload after success of the action. File will be attached as
    ID/public URL to owner.

    """

    def validator(
        key: FlattenKey,
        data: FlattenDataDict,
        errors: FlattenErrorDict,
        context: Context,
    ):
        value: fk.Upload = data.pop(key)
        id_field_path = key[:-1] + (id_field,)
        shared.add_task(
            task.UploadAndAttachTask(
                storage,
                value,
                owner_type,
                id_field_path,
                attach_as,
                using_action,
                destination_field,
            ),
        )

    return validator



"""empty message.

Revision ID: d19094d45a0f
Revises: 6a6ea2472155
Create Date: 2024-06-18 13:23:04.625122

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "d19094d45a0f"
down_revision = "6a6ea2472155"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("files_file", "content_type", nullable=False)
    op.alter_column("files_file", "size", nullable=False)
    op.alter_column("files_file", "hash", nullable=False)


def downgrade():
    op.alter_column("files_file", "hash", nullable=True)
    op.alter_column("files_file", "size", nullable=True)
    op.alter_column("files_file", "content_type", nullable=True)



from __future__ import annotations

import dataclasses
import logging
from typing import Any
from urllib.parse import urlparse, urlunparse

import requests

from ckanext.files import shared

log = logging.getLogger(__name__)


@dataclasses.dataclass()
class Settings(shared.Settings):
    timeout: int = 5


class Reader(shared.Reader):
    capabilities = shared.Capability.PERMANENT_LINK
    storage: LinkStorage

    def permanent_link(self, data: shared.FileData, extras: dict[str, Any]) -> str:
        return data.location


class Uploader(shared.Uploader):
    capabilities = shared.Capability.CREATE
    storage: LinkStorage

    def upload(
        self,
        location: shared.Location,
        upload: shared.Upload,
        extras: dict[str, Any],
    ) -> shared.FileData:
        try:
            url = urlunparse(urlparse(upload.stream.read())).decode()
        except ValueError as err:
            raise shared.exc.ContentError(self, str(err)) from err

        return self.storage.analyze(shared.Location(url))


class Manager(shared.Manager):
    capabilities = shared.Capability.ANALYZE | shared.Capability.REMOVE
    storage: LinkStorage

    def remove(
        self,
        data: shared.FileData | shared.MultipartData,
        extras: dict[str, Any],
    ) -> bool:
        return True

    def analyze(
        self, location: shared.Location, extras: dict[str, Any]
    ) -> shared.FileData:
        resp = requests.head(location, timeout=self.storage.settings.timeout)
        if not resp.ok:
            log.debug("Cannot analyze URL %s: %s", location, resp)

        content_length = resp.headers.get("content-length") or "0"
        size = int(content_length) if content_length.isnumeric() else 0

        content_type = resp.headers.get("content-type") or "application/octet-stream"
        content_type = content_type.split(";", 1)[0]

        hash = resp.headers.get("etag") or ""

        return shared.FileData(location, size, content_type, hash)


class LinkStorage(shared.Storage):
    hidden = True
    settings: Settings  # type: ignore
    SettingsFactory = Settings
    UploaderFactory = Uploader
    ManagerFactory = Manager
    ReaderFactory = Reader

    @classmethod
    def declare_config_options(
        cls,
        declaration: shared.types.Declaration,
        key: shared.types.Key,
    ):
        super().declare_config_options(declaration, key)
        declaration.declare_int(key.timeout, 5).set_description(
            "Request timeout used when link details fetched during ANALYZE",
        )



from __future__ import annotations

import math


def fuw_printable_file_size(size_bytes: int) -> str:
    if size_bytes == 0:
        return "0 bytes"

    size_name = ("bytes", "KB", "MB", "GB", "TB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(float(size_bytes) / p, 1)

    return f"{s} {size_name[i]}"


def fuw_truncate_file_name(file_name: str, max_length: int = 35) -> str:
    """Truncate file name if it is longer than the maximum length.

    Args:
        file_name (str): File name
        max_length (int, optional): Maximum length of the file name. Defaults to 35.

    Returns:
        str: truncated file name
    """
    if len(file_name) <= max_length:
        return file_name

    keep_length = (max_length - 3) // 2
    start = file_name[:keep_length]
    end = file_name[-keep_length:]

    return f"{start}...{end}"
